@@ -437,8 +437,19 @@ void apiClient_invoke(apiClient_t    *apiClient,
                          apiClient);
         curl_easy_setopt(handle, CURLOPT_HTTPHEADER, headers);
         curl_easy_setopt(handle, CURLOPT_VERBOSE, 0); // to get curl debug msg 0: to disable, 1L:to enable
-        curl_easy_setopt(handle, CURLOPT_CONNECTTIMEOUT, 1);
+        const char *timeout_str = getenv("LINGLONG_CONNECT_TIMEOUT");
+        long timeout = 5;
+        if (timeout_str != NULL) {
+            char *endptr;
+            timeout = strtol(timeout_str, &endptr, 10);
 
+            if (endptr == timeout_str || *endptr != '\0') {
+                timeout = 5;
+            } else if (timeout <= 0 || timeout > INT_MAX) {
+                timeout = 5;
+            }
+        }
+        curl_easy_setopt(handle, CURLOPT_CONNECTTIMEOUT, timeout);
 
         if(bodyParameters != NULL) {
             postData(handle, bodyParameters);
